"""Parses sequences in tree to format for DataMonkey FuBar.

This script reads in a *.trees file from BEAST and some FASTA sequence
files giving the coding sequences of the strains in the *.trees file.
It creates a FASTA file with all of the sequences found in the tree, aligned.

Written by Jesse Bloom, 2013."""


import re
import fasta
import align


def main():
    """Main body of script."""
    #
    # input and output files
    intreefile = 'NPhumanH3N2-MarkovJumps_maxcladecredibility.trees'
    inseqfiles = ['Aichi68-NP.fasta', 'BR10-NP.fasta', 'NPhumanH3N2.fasta', 'Nan95-NP.fasta']
    outseqfile = 'seqs.fasta'
    musclepath = '/Users/jbloom/muscle3.8/'
    #
    # parse the tree
    print "\nParsing tree from %s..." % intreefile
    number_to_strain_match = re.compile("^(?P<num>\d+) '(?P<strain>\S+)'\,{0,1}$")
    annotationmatch = re.compile('\[\&[^\]]+\]')
    number_to_strain = {} # defines taxa label numbers
    startedtaxanumbers = foundtree = endedtaxanumbers = False
    for line in open(intreefile):
        line = line.strip()
        if startedtaxanumbers:
            if endedtaxanumbers:
                if line[ : 18] == 'tree TREE1 = [&R] ':
                    tree = line[18 : ].strip()
                    tree = re.sub(annotationmatch, '', tree)
                    foundtree = True
                    break
            elif line == ';':
                endedtaxanumbers = True
            else:
                m = number_to_strain_match.search(line)
                if not m:
                    raise ValueError("Failed to match: %s" % line)
            number_to_strain[int(m.group('num'))] = m.group('strain')
        else:
            if line == 'Translate':
                startedtaxanumbers = True
    assert startedtaxanumbers and endedtaxanumbers and foundtree
    print "Read taxa labels for %d strains." % len(number_to_strain)
    #
    # read the sequences, find those that are in tree, align them
    number_to_seq = {}
    print "\nNow reading in the coding sequences."
    seqs = []
    for f in inseqfiles:
        fseqs = fasta.Read(f)
        print "Read %d sequences from %s." % (len(fseqs), f)
        seqs += fseqs
    print "Read a total of %d sequences from all files." % (len(seqs))
    seqs = [(head.replace(' ', '_'), seq) for (head, seq) in seqs]
    for (number, strain) in number_to_strain.iteritems():
        strainseqs = []
        strain = '_'.join(strain.split('_')[ : -1])
        strain = strain.replace('_variant1', '')
        for (head, seq) in seqs:
            if strain in head:
                number_to_seq[number] = seq
                break
        else:
            raise ValueError("Failed to find sequences for %s" % strain)
    nseqs = len(number_to_seq)
    number_to_seq[1] = number_to_seq[1][ : -3] # remove stop codon
    seqlength = len(number_to_seq[1])
    print "Matched all %d strain numbers to their sequences." % nseqs
    print "\nNow aligning all sequences to the first sequence (length %d)..." % seqlength
    for i in range(2, nseqs + 1):
        a = align.StripGapsToFirstSequence(align.Align([('head1', number_to_seq[1]), ('head2', number_to_seq[i])], musclepath, 'MUSCLE'))[1][1]
        number_to_seq[i] = a
    f = open(outseqfile, 'w')
    for i in range(1, nseqs + 1):
        s = number_to_seq[i]
        assert len(s) == seqlength
        f.write(">%d\n%s\n" % (i, s))
    f.close()
    print "Wrote the aligned sequences to %s." % (outseqfile)


main() # run the script
